#include <stdio.h>
#include <io.h>
#include <errno.h>
#include <sys/stat.h>
/*
_UNICODE UNICODE
or
_MBCS
*/
#include <tchar.h>
#include "std64.h"

/*
errno
E2BIG Argument list too long 7 
EACCES Permission denied 13 
EAGAIN No more processes or not enough memory or maximum nesting level reached 11 
EBADF Bad file number 9 
ECHILD No spawned processes 10 
EDEADLOCK Resource deadlock would occur 36 
EDOM Math argument 33 
EEXIST File exists 17 
EINVAL Invalid argument 22 
EMFILE Too many open files 24 
ENOENT No such file or directory 2 
ENOEXEC Exec format error 8 
ENOMEM Not enough memory 12 
ENOSPC No space left on device 28 
ERANGE Result too large 34 
EXDEV Cross-device link 18 

_doserrno
*/
static void std_findsub(struct std_find_data64 *findbuf)
{
    findbuf->attrib = _A_NORMAL;
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE ) {
        findbuf->attrib |= _A_ARCH;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED ) {
        findbuf->attrib |= _A_CMPRS;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
        findbuf->attrib |= _A_SUBDIR;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED ) {
        findbuf->attrib |= _A_ENCRYPT;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN ) {
        findbuf->attrib |= _A_HIDDEN;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_READONLY ) {
        findbuf->attrib |= _A_RDONLY;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) {
        findbuf->attrib |= _A_RPRS;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_SPARSE_FILE ) {
        findbuf->attrib |= _A_SPRS;
    }
    if ( findbuf->w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM ) {
        findbuf->attrib |= _A_SYSTEM;
    }
    findbuf->time_create = ((ULARGE_INTEGER*)&(findbuf->w32_finddata.ftCreationTime))->QuadPart;
    findbuf->time_access = ((ULARGE_INTEGER*)&(findbuf->w32_finddata.ftLastAccessTime))->QuadPart;
    findbuf->time_write = ((ULARGE_INTEGER*)&(findbuf->w32_finddata.ftLastWriteTime))->QuadPart;
    findbuf->size = ((__int64)(findbuf->w32_finddata.nFileSizeHigh)) * (0x100000000i64) + (__int64)(findbuf->w32_finddata.nFileSizeLow);
    _tcscpy(findbuf->name,findbuf->w32_finddata.cFileName);
}

static void doserrno(void)
{
    _doserrno = GetLastError();
    switch ( _doserrno ) {
      case ERROR_FILE_NOT_FOUND:
      case ERROR_PATH_NOT_FOUND:
      case ERROR_NO_MORE_FILES:
          errno = ENOENT;
          break;
      case ERROR_NOT_ENOUGH_MEMORY:
          errno = ENOMEM;
          break;
      case ERROR_ACCESS_DENIED:
          errno = EACCES;
          break;
      default:
          errno = EINVAL;
          break;
    }
}

int std_findfirst64(_TXCHAR *p,struct std_find_data64 *findbuf)
{
    HANDLE h;

    h = FindFirstFile(p,&findbuf->w32_finddata);
    if ( h == INVALID_HANDLE_VALUE ) {
        doserrno();
        return -1;
    }
    std_findsub(findbuf);
    return (int)h;
}

int std_findnext64(int h,struct std_find_data64 *findbuf)
{
    int sts;

    sts = FindNextFile((HANDLE)h,&findbuf->w32_finddata);
    if ( !sts ) {
        doserrno();
        return -1;
    }
    std_findsub(findbuf);
    return 0;
}

int std_findclose(int h)
{
    if ( !FindClose((HANDLE)h) ) {
        errno = EINVAL;
        return -1;
    }
    return 0;
}

int std_stat64(_TXCHAR *p,struct std_stat64 *sbuf)
{
    struct std_find_data64 findbuf;
    int h;
    DWORD atr;

    h = std_findfirst64(p,&findbuf);
    if ( h == -1 ) {
        atr = GetFileAttributes(p);
        if ( atr == INVALID_FILE_ATTRIBUTES ) {
            return -1;
        }
        sbuf->st_size = -1;
        sbuf->st_atime = -1;
        sbuf->st_mtime = -1;
        sbuf->st_ctime = -1;
    } else {
        atr = findbuf.w32_finddata.dwFileAttributes;
        std_findclose(h);
        sbuf->st_size = findbuf.size;
        sbuf->st_atime = findbuf.time_access;
        sbuf->st_mtime = findbuf.time_write;
        sbuf->st_ctime = findbuf.time_create;
    }
    sbuf->st_dev = -1;
    sbuf->st_nlink = -1;
    sbuf->st_uid = -1;
    sbuf->st_gid = -1;
    sbuf->st_rdev = -1;
    if ( atr & FILE_ATTRIBUTE_DIRECTORY ) {
        sbuf->st_mode = _S_IFDIR;
    } else {
        sbuf->st_mode = _S_IFREG;
    }
    if ( atr & FILE_ATTRIBUTE_READONLY ) {
        sbuf->st_mode |= _S_IREAD;
    } else {
        sbuf->st_mode |= _S_IREAD | _S_IWRITE;
    }
    if ( atr & FILE_ATTRIBUTE_COMPRESSED ) {
        sbuf->st_mode |= _S_IFCMPRS;
    }
    if ( atr & FILE_ATTRIBUTE_ENCRYPTED ) {
        sbuf->st_mode |= _S_IFENCRYPT;
    }
    if ( atr & FILE_ATTRIBUTE_REPARSE_POINT ) {
        sbuf->st_mode |= _S_IFRPRS;
    }
    if ( atr & FILE_ATTRIBUTE_SPARSE_FILE ) {
        sbuf->st_mode |= _S_IFSPRS;
    }
    return 0;
}

static int os; /* 0:no checked  1:ok  -1:fail */

static BOOL (WINAPI *systemtimetotzspecificlocaltime) ( /* NT3.5 or later, not NT3.1 95 98*/ 
  LPTIME_ZONE_INFORMATION lpTimeZoneInformation,
  LPSYSTEMTIME lpUniversalTime,
  LPSYSTEMTIME lpLocalTime
);
static struct std_tm64 tm64;

struct std_tm64* std_localtime64(__int64 *time)
{
    OSVERSIONINFO VersionInfo;
    int sts;
    SYSTEMTIME systemtime;
    SYSTEMTIME localtime;
    FILETIME systemfiletime;
    FILETIME localfiletime;

    if ( os == 0 ) {
        VersionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
        if ( GetVersionEx(&VersionInfo) == 0 ) {
            os = -1;
        } else {
            if ( VersionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT ) {
                systemtimetotzspecificlocaltime = (BOOL (WINAPI *)(LPTIME_ZONE_INFORMATION,LPSYSTEMTIME,LPSYSTEMTIME))GetProcAddress(GetModuleHandle(_T("kernel32.dll")), "SystemTimeToTzSpecificLocalTime");
                if ( systemtimetotzspecificlocaltime != NULL ) {
                    os = 1;
                } else {
                    os = -1;
                }
            }
        }
    }
    systemfiletime.dwLowDateTime = ((ULARGE_INTEGER*)time)->LowPart;
    systemfiletime.dwHighDateTime = ((ULARGE_INTEGER*)time)->HighPart;
    if ( systemtimetotzspecificlocaltime != NULL ) { /* NT3.5 or later, not NT3.1 95 98 */
        sts = FileTimeToSystemTime(&systemfiletime,&systemtime); // UTC FileTime -> SystemTime
        if ( sts == 0 ) {
            _doserrno = GetLastError();
            errno = EINVAL;
            return NULL;
        }
        sts = systemtimetotzspecificlocaltime(NULL,&systemtime,&localtime);
        if ( sts == 0 ) {
            _doserrno = GetLastError();
            errno = EINVAL;
            return NULL;
        }
    } else {
        sts = FileTimeToLocalFileTime(&systemfiletime,&localfiletime); // UTC FileTime -> Local FileTime
        if ( sts == 0 ) {
            _doserrno = GetLastError();
            errno = EINVAL;
            return NULL;
        }
        sts = FileTimeToSystemTime(&localfiletime,&localtime); // Local FileTime -> LocalTime
        if ( sts == 0 ) {
            _doserrno = GetLastError();
            errno = EINVAL;
            return NULL;
        }
    }
    tm64.tm_msec = localtime.wMilliseconds;
    tm64.tm_sec = localtime.wSecond;
    tm64.tm_min = localtime.wMinute;
    tm64.tm_hour = localtime.wHour;
    tm64.tm_mday = localtime.wDay;
    tm64.tm_mon = localtime.wMonth - 1;
    tm64.tm_year = localtime.wYear - 1900;
    tm64.tm_wday = localtime.wDayOfWeek;
    tm64.tm_yday = -1;
    tm64.tm_isdst = -1;
    return &tm64;
}
